#include <opencv2/opencv.hpp>

#include <xmmintrin.h> // SSE intrinsics
#include <emmintrin.h> // SSE2 intrinsics
#include <pmmintrin.h> // SSE3 intrinsics

static void pyrDownMeanSmooth(const cv::Mat& in, cv::Mat& out)
{
  out.create(cv::Size(in.size().width / 2, in.size().height / 2), in.type());

  // aquire pointers to input and output data (OpenCV images are already 16 byte aligned)

  // input data pointer
  const float* in_ptr = in.ptr<float>();
  const float* row1 = in_ptr;
  const float* row2 = in_ptr + in.step1();

  // output data pointer
  float* out_ptr = out.ptr<float>();

  // load constant to register
  __m128 scaling = _mm_set1_ps(0.25);

  __m128 row1_p1, row1_p2, row2_p1, row2_p2, result;

  for(int y = 0; y < out.rows; ++y)
  {
    for(int x = 0; x < out.cols; x += 4, out_ptr += 4, row1 += 8, row2 += 8)
    {
      // load 8 floats from every row into 2 registers (4 in total)
      row1_p1 = _mm_load_ps(row1);
      row1_p2 = _mm_load_ps(row1 + 4);
      row2_p1 = _mm_load_ps(row2);
      row2_p2 = _mm_load_ps(row2 + 4);

      result = _mm_mul_ps(
                  scaling,  
                  _mm_hadd_ps(
                      _mm_add_ps(row1_p1, row2_p1), 
                      _mm_add_ps(row1_p2, row2_p2)
                  )
               );

      // store results to output
      _mm_store_ps(out_ptr, result);

      // alternative: stream directly to RAM without polluting CPU cache, ~10% speedup
      // _mm_stream_ps(out_ptr, result);
    }
  }
}
